@@ -11,41 +11,120 @@
 pub use command_run;
 
 use command_run::Command;
-use std::ffi::OsString;
-use std::path::{Path, PathBuf};
+use std::ffi::{OsStr, OsString};
+use std::path::PathBuf;
+use std::{env, fmt};
+
+/// Preset base commands that a [`Launcher`] can be constructed from.
+#[derive(Clone, Copy, Debug, Eq, PartialEq)]
+pub enum BaseCommand {
+    /// Docker without sudo.
+    Docker,
+
+    /// Docker with sudo.
+    SudoDocker,
+
+    /// Podman.
+    Podman,
+}
+
+// TODO: is there a good existing crate for this? I found a few on
+// crates.io that didn't look quite right.
+fn is_exe_in_path(exe_name: &OsStr) -> bool {
+    let paths = if let Some(paths) = env::var_os("PATH") {
+        paths
+    } else {
+        return false;
+    };
+
+    env::split_paths(&paths).any(|path| path.join(exe_name).exists())
+}
+
+// TODO: consider using nix or some other crate.
+fn is_user_in_group(target_group: &str) -> bool {
+    let mut cmd = Command::new("groups");
+    cmd.log_command = false;
+    cmd.capture = true;
+    cmd.log_output_on_error = true;
+    let output = if let Ok(output) = cmd.run() {
+        output
+    } else {
+        return false;
+    };
+    let stdout = output.stdout_string_lossy();
+    stdout.split_whitespace().any(|group| group == target_group)
+}
 
 /// Base container command used for building and running containers.
 ///
 /// This allows variations such as "docker", "sudo docker", and
 /// "podman".
 #[derive(Clone, Debug, Eq, PartialEq)]
-pub struct Docker {
-    /// If true, run the command with `sudo`. Defaults to false.
-    pub sudo: bool,
-
-    /// The container command. The default is "docker", but can be
-    /// changed to another compatible program such as "podman".
-    pub program: PathBuf,
+pub struct Launcher {
+    base_command: Command,
 }
 
-impl Docker {
-    /// Create a new Docker instance with the default values set.
-    pub fn new() -> Self {
-        Self::default()
+impl Launcher {
+    /// Create a new `Launcher` with the specified base [`Command`]. The
+    /// base command is used to create all the other commands.
+    pub fn new(base_command: Command) -> Self {
+        Self { base_command }
     }
 
-    /// Create the base `Command` for running Docker.
-    pub fn command(&self) -> Command {
-        if self.sudo {
-            Command::with_args("sudo", &[&self.program])
+    /// Automatically choose a base command.
+    ///
+    /// * Chooses `podman` if is in the `$PATH`.
+    /// * Otherwise chooses `docker` if it is in the `$PATH`.
+    ///   * If the current user is not in a `docker` group, `sudo` is added.
+    ///
+    /// If neither command is in the `$PATH`, returns `None`.
+    pub fn auto() -> Option<Self> {
+        let docker = OsStr::new("docker");
+        let podman = OsStr::new("podman");
+        if is_exe_in_path(podman) {
+            Some(BaseCommand::Podman.into())
+        } else if is_exe_in_path(docker) {
+            Some(if is_user_in_group("docker") {
+                BaseCommand::Docker.into()
+            } else {
+                BaseCommand::SudoDocker.into()
+            })
         } else {
-            Command::new(&self.program)
+            None
         }
     }
 
-    /// Create a `Command` for building a container.
+    /// Whether the base command appears to be the given `program`. This
+    /// checks if base command's program or any arguments match the
+    /// input `program`.
+    fn is_program(&self, program: &str) -> bool {
+        let program = OsStr::new(program);
+        self.base_command.program == program
+            || self.base_command.args.contains(&program.into())
+    }
+
+    /// Whether the base command appears to be docker. This checks if
+    /// the program or any of the arguments in the base command match
+    /// "docker".
+    pub fn is_docker(&self) -> bool {
+        self.is_program("docker")
+    }
+
+    /// Whether the base command appears to be podman. This checks if
+    /// the program or any of the arguments in the base command match
+    /// "podman".
+    pub fn is_podman(&self) -> bool {
+        self.is_program("podman")
+    }
+
+    /// Get the base [`Command`].
+    pub fn base_command(&self) -> &Command {
+        &self.base_command
+    }
+
+    /// Create a [`Command`] for building a container.
     pub fn build(&self, opt: BuildOpt) -> Command {
-        let mut cmd = self.command();
+        let mut cmd = self.base_command.clone();
         cmd.add_arg("build");
 
         // --build-arg
@@ -82,27 +161,27 @@ impl Docker {
         cmd
     }
 
-    /// Create a `Command` for creating a network.
+    /// Create a [`Command`] for creating a network.
     pub fn create_network(&self, opt: CreateNetworkOpt) -> Command {
-        let mut cmd = self.command();
+        let mut cmd = self.base_command.clone();
         cmd.add_arg_pair("network", "create");
         cmd.add_arg(opt.name);
 
         cmd
     }
 
-    /// Create a `Command` for removing a network.
+    /// Create a [`Command`] for removing a network.
     pub fn remove_network(&self, name: &str) -> Command {
-        let mut cmd = self.command();
+        let mut cmd = self.base_command.clone();
         cmd.add_arg_pair("network", "rm");
         cmd.add_arg(name);
 
         cmd
     }
 
-    /// Create a `Command` for running a container.
+    /// Create a [`Command`] for running a container.
     pub fn run(&self, opt: RunOpt) -> Command {
-        let mut cmd = self.command();
+        let mut cmd = self.base_command.clone();
         cmd.add_arg("run");
 
         // --detach
@@ -164,11 +243,18 @@ impl Docker {
     }
 }
 
-impl Default for Docker {
-    fn default() -> Self {
-        Docker {
-            sudo: false,
-            program: Path::new("docker").into(),
+impl From<BaseCommand> for Launcher {
+    fn from(bc: BaseCommand) -> Launcher {
+        let docker = "docker";
+        let podman = "podman";
+        Self {
+            base_command: match bc {
+                BaseCommand::Docker => Command::new(docker),
+                BaseCommand::SudoDocker => {
+                    Command::with_args("sudo", &[docker])
+                }
+                BaseCommand::Podman => Command::new(podman),
+            },
         }
     }
 }
@@ -218,41 +304,61 @@ pub enum NameOrId {
     Id(u32),
 }
 
-impl NameOrId {
-    /// Format as an argument.
-    pub fn arg(&self) -> String {
+impl fmt::Display for NameOrId {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match self {
-            NameOrId::Name(name) => name.clone(),
-            NameOrId::Id(id) => id.to_string(),
+            Self::Name(name) => write!(f, "{}", name),
+            Self::Id(id) => write!(f, "{}", id),
         }
     }
 }
 
-/// User specification used when running a container.
+impl From<String> for NameOrId {
+    fn from(name: String) -> Self {
+        Self::Name(name)
+    }
+}
+
+impl From<u32> for NameOrId {
+    fn from(id: u32) -> Self {
+        Self::Id(id)
+    }
+}
+
+/// User and (optionally) group.
 #[derive(Clone, Debug, Eq, PartialEq)]
-pub struct User {
-    /// User or UID
+pub struct UserAndGroup {
+    /// User or UID.
     pub user: NameOrId,
 
-    /// Group or GID
+    /// Group or GID.
     pub group: Option<NameOrId>,
 }
 
-impl User {
-    /// Get a `User` with the current UID and GID set.
-    pub fn current() -> User {
-        User {
-            user: NameOrId::Id(users::get_current_uid()),
-            group: Some(NameOrId::Id(users::get_current_gid())),
+impl UserAndGroup {
+    /// Get a `UserAndGroup` with the current UID and GID set.
+    pub fn current() -> Self {
+        Self {
+            user: users::get_current_uid().into(),
+            group: Some(users::get_current_gid().into()),
         }
     }
 
-    /// Format as an argument.
+    /// Get a `UserAndGroup` with UID and GID set to zero.
+    pub fn root() -> Self {
+        Self {
+            user: 0.into(),
+            group: Some(0.into()),
+        }
+    }
+
+    /// Format as an argument. If `group` is set, the format is
+    /// `<user>:<group>`, otherwise just `<user>`.
     pub fn arg(&self) -> String {
-        let mut out = self.user.arg();
+        let mut out = self.user.to_string();
         if let Some(group) = &self.group {
             out.push(':');
-            out.push_str(&group.arg());
+            out.push_str(&group.to_string());
         }
         out
     }
@@ -319,8 +425,8 @@ pub struct RunOpt {
     /// Connect a container to a network.
     pub network: Option<String>,
 
-    /// User (and optionally group) to use inside the container.
-    pub user: Option<User>,
+    /// User (and optionally) group to use inside the container.
+    pub user: Option<UserAndGroup>,
 
     /// Mount the container's root filesystem as read only.
     pub read_only: bool,
@@ -338,20 +444,3 @@ pub struct RunOpt {
     /// Optional arguments to pass to the command.
     pub args: Vec<OsString>,
 }
-
-#[cfg(test)]
-mod tests {
-    use super::*;
-
-    #[test]
-    fn test_command() {
-        let mut docker = Docker::new();
-        assert_eq!(docker.command().command_line_lossy(), "docker");
-
-        docker.sudo = true;
-        assert_eq!(docker.command().command_line_lossy(), "sudo docker");
-
-        docker.program = "myCommand".into();
-        assert_eq!(docker.command().command_line_lossy(), "sudo myCommand");
-    }
-}
